/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright  2000-2003 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/
/**
    \file 
        Caelum.h
    \brief
        Test for the caelum plugin 
*/

#include "ExampleApplication.h"
#include "OgreStringConverter.h"

#include "CaelumMain.h"

#define FLOW_SPEED 0.2
#define FLOW_HEIGHT 0.8

class CaelumListener : public ExampleFrameListener {
  public:
    CaelumListener (SceneManager *sceneMgr, RenderWindow* win, Camera* cam, CaelumManager *caelumManager) : ExampleFrameListener(win, cam) {
      mSceneMgr = sceneMgr;
      mCaelumMgr = caelumManager;

      mNumScreenShots = 0;

      mElapsedTime = 0;
    };

    void updateCaelumStats () {
      static String currTime = "Current time: ";
      static String currLocation = "Current location: ";

      // update stats when necessary
      GuiElement *guiCurrTime = GuiManager::getSingleton().getGuiElement("Caelum/CurrTime");
      GuiElement *guiCurrLocation = GuiManager::getSingleton().getGuiElement("Caelum/CurrLocation");

      Real time = mCaelumMgr->getTime ();
      guiCurrTime->setCaption(currTime + StringConverter::toString(((unsigned int )(time)) / 3600) 
        + ":" + StringConverter::toString((((unsigned int )(time)) % 3600) / 60) 
        + ":" + StringConverter::toString((((unsigned int )(time)) % 60))
        + "." + StringConverter::toString((unsigned int )((time - ((unsigned int )(time))) * 100)));
      Vector3 loc = mCaelumMgr->getLocation ();
      String latitude, longitude;
      if (loc.x >= 0)
        longitude = StringConverter::toString (loc.x) + "E";
      else
        longitude = StringConverter::toString (Math::Abs (loc.x)) + "W";
      if (loc.y >= 0)
        latitude = StringConverter::toString (loc.y) + "N";
      else
        latitude = StringConverter::toString (Math::Abs (loc.y)) + "S";
      guiCurrLocation->setCaption(currLocation + latitude + " " + longitude); 
    }

 // Override frameStarted event to process that (don't care about frameEnded)
    bool frameStarted (const FrameEvent &evt) {
      float moveScale;
      float rotScale;
      float waterFlow;
      static float flowAmount = 0.0f;
      static bool flowUp = true;
    	static bool wireframe = false;
	    static Real targetQuality = 3.0, minimumQuality = 10;

      // Update elapsed time
      mElapsedTime += evt.timeSinceLastFrame;

      // local just to stop toggles flipping too fast
      static Real timeUntilNextToggle = 0;

      if (timeUntilNextToggle >= 0)
        timeUntilNextToggle -= evt.timeSinceLastFrame;

      // If this is the first frame, pick a speed
      if (evt.timeSinceLastFrame == 0) {
        moveScale = 1;
        rotScale = 0.1;
        waterFlow = 0.0f;
      }
      // Otherwise scale movement units by time passed since last frame
      else {
        // Move about 2 units per second,
        moveScale = 50 * evt.timeSinceLastFrame;
        // Take about 10 seconds for full rotation
        rotScale = 36 * evt.timeSinceLastFrame;
            
        // set a nice waterflow rate
        waterFlow = FLOW_SPEED * evt.timeSinceLastFrame;            
      }

      // Grab input device state
      mInputDevice->capture ();

      SceneNode *waterNode = static_cast<SceneNode*>(mCamera->getSceneManager ()->getRootSceneNode ()->getChild ("WaterNode"));
      if (waterNode) {
        if (flowUp)
          flowAmount += waterFlow;
        else
          flowAmount -= waterFlow;

        if (flowAmount >= FLOW_HEIGHT)
          flowUp = false;
        else if (flowAmount <= 0.0f)
          flowUp = true;

        waterNode->translate (0, (flowUp ? waterFlow : -waterFlow), 0);
      }

      static Vector3 vec;

      vec = Vector3::ZERO;

      if (mInputDevice->isKeyDown(KC_A)) {
        // Move camera left
        vec.x = -moveScale;
      }

      if (mInputDevice->isKeyDown(KC_D)) {
        // Move camera RIGHT
        vec.x = moveScale;
      }

      if (mInputDevice->isKeyDown(KC_UP) || mInputDevice->isKeyDown(KC_W)) {
        // Move camera forward
        vec.z = -moveScale;
      }

      if (mInputDevice->isKeyDown(KC_DOWN) || mInputDevice->isKeyDown(KC_S)) {
        // Move camera backward
        vec.z = moveScale;
      }

      if (mInputDevice->isKeyDown(KC_PGUP)) {
        // Move camera up
        vec.y = moveScale;
      }

      if (mInputDevice->isKeyDown(KC_PGDOWN)) {
        // Move camera down
        vec.y = -moveScale;
      }

      if (mInputDevice->isKeyDown(KC_RIGHT)) {
        mCamera->yaw(-rotScale);
      }
      if (mInputDevice->isKeyDown(KC_LEFT)) {
        mCamera->yaw(rotScale);
      }

      if (mInputDevice->isKeyDown( KC_ESCAPE)) {            
        return false;
      }

      // Rotate view by mouse relative position
      float rotX, rotY;
      rotX = -mInputDevice->getMouseRelativeX() * 0.13;
      rotY = -mInputDevice->getMouseRelativeY() * 0.13;


      // Make all the changes to the camera
      // Note that YAW direction is around a fixed axis (freelook stylee) rather than a natural YAW (e.g. airplane)
      mCamera->yaw(rotX);
      mCamera->pitch(rotY);
      mCamera->moveRelative(vec);

      // Rotate scene node if required
      SceneNode *node = mCamera->getSceneManager ()->getRootSceneNode ();
      if (mInputDevice->isKeyDown(KC_O)) {
        node->yaw(rotScale);
      }
      if (mInputDevice->isKeyDown(KC_P)) {
        node->yaw(-rotScale);
      }
      if (mInputDevice->isKeyDown(KC_I)) {
        node->pitch(rotScale);
      }
      if (mInputDevice->isKeyDown(KC_K)) {
        node->pitch(-rotScale);
      }

      if (mInputDevice->isKeyDown(KC_F) && timeUntilNextToggle <= 0) {
        mStatsOn = !mStatsOn;
        showDebugOverlay(mStatsOn);
        if (static_cast<Overlay *>(OverlayManager::getSingleton ().getByName ("Caelum/InfoOverlay"))->isVisible ())
          static_cast<Overlay *>(OverlayManager::getSingleton ().getByName ("Caelum/InfoOverlay"))->hide ();
        else
          static_cast<Overlay *>(OverlayManager::getSingleton ().getByName ("Caelum/InfoOverlay"))->show ();

        timeUntilNextToggle = 1;
      }

      if (mInputDevice->isKeyDown(KC_T) && timeUntilNextToggle <= 0) {
  	    wireframe = !wireframe;
	      if (wireframe)
		      mCamera->setDetailLevel(SDL_WIREFRAME);
  	    else
	      	mCamera->setDetailLevel(SDL_SOLID);

        timeUntilNextToggle = 0.5;
      }

    	if (mInputDevice->isKeyDown(KC_9) && timeUntilNextToggle <= 0) {
  	    targetQuality += 0.25;
	      mSceneMgr->setOption("TARGET_QUALITY", &targetQuality);
	      timeUntilNextToggle = 0.20;
	    }

	    if (mInputDevice->isKeyDown(KC_8) && timeUntilNextToggle <= 0) {
  	    targetQuality -= 0.25;
	      mSceneMgr->setOption("TARGET_QUALITY", &targetQuality);
	      timeUntilNextToggle = 0.20;
	    }

    	if (mInputDevice->isKeyDown(KC_6) && timeUntilNextToggle <= 0) {
  	    minimumQuality += 1;
	      mSceneMgr->setOption("MINIMUM_QUALITY", &minimumQuality);
	      timeUntilNextToggle = 0.20;
	    }

  	  if (mInputDevice->isKeyDown(KC_5) && timeUntilNextToggle <= 0) {
  	    minimumQuality -= 1;
	      mSceneMgr->setOption("MINIMUM_QUALITY", &minimumQuality);
	      timeUntilNextToggle = 0.20;
	    }

      if (mInputDevice->isKeyDown (KC_SYSRQ) && timeUntilNextToggle <= 0) {
			  char tmp[20];
			  sprintf(tmp, "screenshot_%d.png", ++mNumScreenShots);
        mWindow->writeContentsToFile(tmp);
        timeUntilNextToggle = 0.2;
  			mWindow->setDebugText(String("Wrote ") + tmp);
      }

      // Update caelum
      mCaelumMgr->update (evt.timeSinceLastFrame * 800);

      updateCaelumStats ();

      // Return true to continue rendering
      return true;
    }

  private:
    SceneManager *mSceneMgr;
    CaelumManager *mCaelumMgr;

    unsigned int mNumScreenShots;

    Real mElapsedTime;
};


class CaelumApplication : public ExampleApplication {
  public:
    CaelumApplication() { }

  protected:
    // CAELUM!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    CaelumManager *mCaelumMgr;
    SceneNode *mSkyDome;

    virtual void createFrameListener(void) {
      mFrameListener = new CaelumListener (mSceneMgr, mWindow, mCamera, mCaelumMgr);
      mFrameListener->showDebugOverlay (true);
      mRoot->addFrameListener (mFrameListener);
    }


    virtual void chooseSceneManager(void) {
      // Get the SceneManager, in this case a generic one
      mSceneMgr = mRoot->getSceneManager( ST_EXTERIOR_FAR );
    }

    virtual void createCamera(void) {
      // Create the camera
      mCamera = mSceneMgr->createCamera("PlayerCam");

      // Position it at 500 in Z direction
      mCamera->setPosition(Vector3(0, 0, 0));
      // Look towards north
      mCamera->lookAt(Vector3(0,0,1));
      mCamera->setNearClipDistance( .5 );

       mCamera->setFarClipDistance( 5000 );
    }
   
    virtual void createViewports(void) {
      // Create one viewport, entire window
      Viewport* vp = mWindow->addViewport(mCamera);
      vp->setBackgroundColour(ColourValue(0.77, 0.86, 1.0));
    }

    // Just override the mandatory create scene method
    void createScene(void) {
      Entity *waterEntity;
      Plane waterPlane;

      // First of all, instantiate Caelum
      mCaelumMgr = new CaelumManager (mSceneMgr, mCamera);
        // Just to try some random data... ;)
        mCaelumMgr->setLocation (Math::RangeRandom (-10000, 10000), Math::RangeRandom (-10000, 10000));
//        mCaelumMgr->setTime (Math::RangeRandom (0, 23), Math::RangeRandom (0, 59), Math::RangeRandom (0, 59));
        // Real time -- REAL, SUMMER TIME!!!
        time_t localTime;
        time (&localTime);
        localTime %= 86400; // To time (supress date)
        mCaelumMgr->setTime (localTime / 3600, (localTime % 3600) / 60, localTime % 60);
      //mCaelumMgr->getHaze ()->setTurbidity (0.0005);
      Image *img = new Image ();
      img->load ("hazeColours.png");
      mCaelumMgr->getHaze ()->setColourMap (img);
      mCaelumMgr->getDome ()->setPressure (5);
      Image *img2 = new Image ();
      img2->load ("SkyDome.png");
      mCaelumMgr->getDome ()->setColourMap (img2);
      BillboardSet *bbs = mSceneMgr->createBillboardSet ("SunObjectsBBS");
      bbs->createBillboard (0, 0, 0);
      bbs->setMaterialName ("Examples/Flare3");
      bbs->setDefaultDimensions (200, 200);
      mCaelumMgr->createObject ("BizarreSun", bbs);
      CaelumObject *obj = mCaelumMgr->getObject ("BizarreSun");
      Image *img3 = new Image ();
      img3->load ("SunColours.png");
      obj->setColourMap (img3);
      obj->setTimeOffset (-.25);

/*
      // Stormy day :)
      mCaelumMgr->getHaze ()->setSaturationModifier (0.3);
      mCaelumMgr->getDome ()->setSaturationModifier (0.3);
      mCaelumMgr->getHaze ()->setColourModifier (ColourValue (0.2, 0.3, 0.5));
      mCaelumMgr->getDome ()->setColourModifier (ColourValue (0.8, 0.85, 1));
*/

      // create a water plane/scene node
      waterPlane.normal = Vector3::UNIT_Y; 
      waterPlane.d = -30; 
      MeshManager::getSingleton().createPlane("WaterPlane", waterPlane, 2800, 2800, 20, 20, true, 1, 10, 10, Vector3::UNIT_Z);
      waterEntity = mSceneMgr->createEntity("water", "WaterPlane"); 
      waterEntity->setMaterialName("Examples/TextureEffect4"); 

      SceneNode *waterNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("WaterNode"); 
      waterNode->attachObject(waterEntity); 
      waterNode->translate(1000, 0, 1000);

      // Create a light
//      Light* l = mSceneMgr->createLight("MainLight");
      // Accept default settings: point light, white diffuse, just set position
      // NB I could attach the light to a SceneNode if I wanted it to move automatically with
      //  other objects, but I don't
//      l->setPosition(20,80,50);

      mSceneMgr->setWorldGeometry( "world.cfg" );

      mSceneMgr->setFog(FOG_EXP2, ColourValue(0.77, 0.86, 1.0), 0.0015, 0,0 );
      //mSceneMgr->setFog(FOG_LINEAR, ColourValue(0.77, 0.86, 1.0), 0, 150, 500);
      //mRoot -> showDebugOverlay( true );

      mCamera->setPosition(0,100,0);
    }
};
